.navbar {
    display: none;
    border: none;
}

.infobar {
    display: none;
    border: none;
}

.visualisation {
    position:absolute;
    border: 1px solid #aaa;
    border-radius: 10px;
    padding: 5px;
    width: 100%;
    height: 100%;
    background-color: white;
    border: none;
}

.size_box {
    z-index:1;
    position:absolute;
    top: 10px;
    left:10px;
    border: 1px solid steelblue;
    border-radius: 5px;
    padding: 5px;
    width: 50px%;
    height: 50px%;
    background-color: white;
    display: inline;
}

path.link {
    fill: none;
    stroke: steelblue;
}

